import json
from abc import ABC, abstractmethod

from osol.extremum.tools.coding import DecodeToNumpy, EncodeFromNumpy


class TerminationException(Exception):
    pass


class OptimizationAlgorithm(ABC):
    @abstractmethod
    def initialize(self, f, search_area):
        raise NotImplementedError()

    @abstractmethod
    def iterate(self, f, search_area):
        raise NotImplementedError()

    @abstractmethod
    def terminate(self, f, search_area):
        raise NotImplementedError()

    def serialize(self):
        return json.dumps(self.__dict__, indent=4, cls=EncodeFromNumpy)

    def deserialize(self, state):
        self.__dict__ = json.loads(state, cls=DecodeToNumpy)

    def optimize(self, f, search_area, number_of_iterations, **kwargs):
        if not kwargs.get("skip_init", False):
            self.initialize(f, search_area)
        for _ in range(number_of_iterations):
            try:
                self.iterate(f, search_area)
            except TerminationException:
                break
        solution = self.terminate(f, search_area)
        return solution



import numpy as np


def smoke_check(algorithm, number_of_iterations):
    left, right = -10, 10
    search_area = np.array([[left, right], [left, right]])
    x_opt = np.random.uniform(left, right, 2)
    eps = 1e-2
    x_found = algorithm.optimize(
        lambda x: np.sum((x - x_opt) ** 2), search_area, number_of_iterations
    )
    return np.linalg.norm(x_found - x_opt) < eps



import json

import numpy


class EncodeFromNumpy(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, numpy.ndarray):
            return {"_kind_": "ndarray", "_value_": o.tolist()}
        return super().default(o)


class DecodeToNumpy(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(
            self, object_hook=DecodeToNumpy.decoder_hook, *args, **kwargs
        )

    @staticmethod
    def decoder_hook(obj):
        if "_kind_" not in obj:
            return obj
        kind = obj["_kind_"]
        if kind == "ndarray":
            return numpy.array(obj["_value_"])
        return obj



from osol.extremum.algorithm import OptimizationAlgorithm
from osol.extremum.tools.vectors import (
    bound_vector,
    generate_vector_in_area,
    generate_vector_in_sphere,
)


class RandomSearch(OptimizationAlgorithm):
    def __init__(self, radius):
        self.radius = radius

    def initialize(self, f, search_area):
        self.x = generate_vector_in_area(search_area)
        self.f_x = f(self.x)

    def iterate(self, f, search_area):
        x_new = generate_vector_in_sphere(self.x, self.radius)
        x_new = bound_vector(x_new, search_area)
        f_x_new = f(x_new)
        if f_x_new < self.f_x:
            self.x = x_new
            self.f_x = f_x_new

    def terminate(self, _, __):
        return self.x



from osol.extremum.algorithm import OptimizationAlgorithm
from osol.extremum.tools.vectors import (
    bound_vector,
    generate_vector_in_area,
    generate_vector_in_sphere,
)


class LuusJaakola(OptimizationAlgorithm):
    def __init__(
        self,
        init_radius,
        number_of_samples,
        reduction_coefficient,
        recover_coefficient,
        iteration_per_run,
    ):
        self.init_radius = self._r = init_radius
        self.number_of_samples = number_of_samples
        self.reduction_coefficient = reduction_coefficient
        self.recover_coefficient = recover_coefficient
        self.iteration_per_run = iteration_per_run

        self._run_id = 0
        self._iter_id = 0

    def initialize(self, f, search_area):
        self.x = generate_vector_in_area(search_area)
        self.f_x = f(self.x)

    def iterate(self, f, search_area):
        for _ in range(self.number_of_samples):
            x_new = generate_vector_in_sphere(self.x, self._r)
            x_new = bound_vector(x_new, search_area)
            f_x_new = f(x_new)
            if f_x_new < self.f_x:
                self.x = x_new
                self.f_x = f_x_new

        self._r *= self.reduction_coefficient
        self._iter_id += 1

        if self._iter_id == self.iteration_per_run:
            self._iter_id = 0
            self._run_id += 1
            self._r = (self.recover_coefficient ** self._run_id) * self.init_radius

    def terminate(self, _, __):
        return self.x



import numpy as np

from osol.extremum.algorithm import OptimizationAlgorithm
from osol.extremum.tools.vectors import (
    bound_vector,
    generate_vector_in_area,
    generate_vector_in_sphere,
)


class StatisticalAntiGradientRandomSearch(OptimizationAlgorithm):
    def __init__(self, radius, number_of_samples):
        self.radius = radius
        self.number_of_samples = number_of_samples

    def initialize(self, f, search_area):
        self.x = generate_vector_in_area(search_area)
        self.f_x = f(self.x)

    def iterate(self, f, search_area):
        antigrad_x = [
            generate_vector_in_sphere(self.x, self.radius)
            for _ in range(self.number_of_samples)
        ]
        antigrad_f_x = [f(x) for x in antigrad_x]

        anti_gradient = np.zeros(shape=len(search_area))
        for temp_x, temp_f_x in zip(antigrad_x, antigrad_f_x):
            anti_gradient -= (temp_x - self.x) * (temp_f_x - self.f_x)
        anti_gradient_length = np.linalg.norm(anti_gradient)
        if anti_gradient_length > 0:
            anti_gradient /= anti_gradient_length

        x_new = self.x + np.random.uniform(0.0, self.radius) * anti_gradient
        x_new = bound_vector(x_new, search_area)
        f_x_new = f(x_new)
        if f_x_new < self.f_x:
            self.x = x_new
            self.f_x = f_x_new

    def terminate(self, _, __):
        return self.x
